import numpy as np


def bin_stats(arr) -> str:
    """Return a string basic statistics for a binary array of either {0,1} or {True,False}"""
    n = len(arr)
    k = sum(arr)
    p = 100. * k / n if n != 0 else 0.
    return f"{k}/{n} ({p:.2f}%)"


def to_python_type(v, ndigits: int):
    if v is None or isinstance(v, str) or isinstance(v, int):
        return v
    if isinstance(v, float):
        v = round(v, ndigits)
        return int(v) if v.is_integer() else v

    if isinstance(v, list):
        return [to_python_type(x, ndigits) for x in v]
    if isinstance(v, np.ndarray):
        return v.round(ndigits).tolist()
    if isinstance(v, np.number):
        return to_python_type(v.item(), ndigits)

    raise ValueError(f"Unknown type for {v}: {type(v)}")



import functools
import itertools
import logging
import os
import pickle
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import List

_logger = logging.getLogger(__name__)


def flatten(collection):
    return list(itertools.chain(*collection))


# copied from maskrcnn-benchmark
class HidePrints:
    """Context Manager that mutes sys.stdout so that print() statements have no effect"""

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def set_gpu_ids(gpu_ids: List[int]):
    gpu_ids_str = ",".join(map(str, gpu_ids))
    os.environ['CUDA_VISIBLE_DEVICES'] = gpu_ids_str


def pickle_cache(cache_path: Path):
    """
    Decorator to cache the result of a function in a pickle file.
    :param cache_path: the path where to cache the file
    """

    def decorator(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            if cache_path.is_file():
                with cache_path.open("rb") as f:
                    return pickle.load(f)
            res = func(*args, **kwargs)
            _logger.info("Caching object type %s to %s", type(res), cache_path)
            cache_path.parent.mkdir(parents=True, exist_ok=True)
            with cache_path.open("wb") as f:
                pickle.dump(res, f)
            return res

        return inner

    return decorator


@contextmanager
def cwd(path):
    oldpwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(oldpwd)


@contextmanager
def change_log_level(logger: logging.Logger, level):
    old_level = logger.level
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(old_level)



import logging
from typing import Union, List

import numpy as np
import pandas as pd
import torch

logger = logging.getLogger(__name__)


def regression_metrics(y_true: Union[np.ndarray, torch.Tensor],
                       y_pred: Union[np.ndarray, torch.Tensor],
                       target_names: List[str]):
    """
    Regression metric calculation that supports both multiple targets and single targets scenario.
    Result can be visualized e.g. using `metrics[["mse", "mae"]].plot.bar(figsize=(8,4));`
    :param y_true: true
    :param y_pred: predictions
    :param target_names a list of target names
    :return: a pandas DataFrame with metrics as columns and target names as rows
    """
    if isinstance(y_true, torch.Tensor):
        y_true = y_true.cpu().numpy()
    if isinstance(y_pred, torch.Tensor):
        y_pred = y_pred.cpu().numpy()
    y_true_df = pd.DataFrame(y_true, columns=target_names)
    y_pred_df = pd.DataFrame(y_pred, columns=target_names)
    return _regression_metrics(y_true_df, y_pred_df)


def regression_metrics_pandas(y_true: pd.DataFrame, y_pred: pd.DataFrame) -> pd.DataFrame:
    """
    Regression metric calculation that supports both multiple targets and single targets scenario.
    Result can be visualized e.g. using `metrics[["mse", "mae"]].plot.bar(figsize=(8,4));`
    :param y_true: true
    :param y_pred: predictions
    :return: a pandas DataFrame with metrics as columns and target names as rows
    """
    return _regression_metrics(y_true, y_pred)


def _regression_metrics(y_true: Union[pd.DataFrame, pd.Series, np.ndarray],
                        y_pred: Union[pd.DataFrame, pd.Series, np.ndarray],
                        name: str = "") -> pd.DataFrame:
    if y_true.ndim > 1 and y_true.shape[1] > 1:  # multiple output scenario
        # recursively call function for each column and combine results
        return pd.concat(_regression_metrics(y_true[col], y_pred[col], name=col) for col in y_true.columns)

    # extract values if necessary
    if isinstance(y_true, pd.DataFrame) and y_true.shape[1] == 1:
        y_true = y_true.iloc[:, 0]
    if isinstance(y_pred, pd.DataFrame) and y_pred.shape[1] == 1:
        y_pred = y_pred.iloc[:, 0]
    if isinstance(y_true, pd.Series):
        y_true = y_true.values
    if isinstance(y_pred, pd.Series):
        y_pred = y_pred.values

    res_df = pd.DataFrame({
        'target': name,
        'y_true': y_true,
        'y_pred': y_pred,
        'ae': np.abs(y_true - y_pred),
        'se': (y_true - y_pred) ** 2
    })
    res_df['ape'] = (res_df['ae'] / res_df.y_true) * 100.
    mse = np.mean(res_df.se)
    mae = np.mean(res_df.ae)
    mape = np.mean(res_df.ape)
    median_ae = np.median(res_df.se)
    return pd.Series(dict(mse=mse, mae=mae, mape=mape, median_ae=median_ae), name=name).to_frame().T



import sys

from yamlu.img import read_img, plot_img, plot_imgs, plot_img_paths, plot_anns
from yamlu.misc import flatten
from yamlu.np_utils import bin_stats
from yamlu.path import ls, glob

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

all = [
    "read_img", "plot_img", "plot_imgs", "plot_img_paths", "plot_anns",
    "flatten",
    "bin_stats",
    "ls", "glob",
]

try:
    from yamlu.pytorch import isin

    all.append("isin")
except ImportError:
    pass

__all__ = all



import json
from collections import defaultdict
from pathlib import Path
from typing import List, Set

import numpy as np
from tqdm import tqdm

from yamlu.img import AnnotatedImage, Annotation, BoundingBox

ARROW_RELATIONS = ["arrow_prev", "arrow_next"]


class CocoReader:
    """
    Parses diagram datasets stored in COCO format (https://cocodataset.org/#format-data).
    The COCO format is extended with some special relation fields for arrows and text.
    """

    def __init__(self, dataset_root: Path, arrow_categories: Set[str] = ("arrow",)):
        self.dataset_root = dataset_root
        self.arrow_categories = set(arrow_categories)
        self.splits = [p.name for p in self.dataset_root.iterdir() if p.is_dir() and not p.name.startswith(".")]

    def parse_split(self, split: str) -> List[AnnotatedImage]:
        json_path = self.dataset_root / f"{split}.json"
        imgs_path = self.dataset_root / split

        coco = json.loads(json_path.read_text())
        image_id_to_anns = defaultdict(list)
        for d in coco["annotations"]:
            image_id_to_anns[d["image_id"]].append(d)

        ais = []
        for coco_image in tqdm(coco["images"]):
            coco_anns = image_id_to_anns[coco_image["id"]]
            anns = self._parse_coco_anns(coco_anns)
            img_path = imgs_path / coco_image["file_name"]
            ai = AnnotatedImage.from_img_path(img_path, anns)
            ais.append(ai)

        return ais

    def _parse_coco_anns(self, coco_anns) -> List[Annotation]:
        anns = []
        for coco_ann in coco_anns:
            a = Annotation(
                category=coco_ann["category"],
                bb=BoundingBox.from_xywh(*coco_ann["bbox"]),
                id=coco_ann["id"],
            )
            # parse keypoints for arrows
            # In COCO format, "keypoints" is a length 3k array
            # Each keypoint has a location x,y and a visibility flag v defined as:
            # - v=0: not labeled (in which case x=y=0)
            # - v=1: labeled but not visible
            # - v=2: labeled and visible.
            # In the diagram datasets all arrow keypoints are v=2 "labeled and visible", so we can ignore this flag
            if a.category in self.arrow_categories:
                a.keypoints = np.array(coco_ann["keypoints"]).reshape(-1, 3)
            anns.append(a)

        id_to_ann = {a.id: a for a in anns}

        # parse arrow source and target relations
        coco_arrows = [a for a in coco_anns if a["category"] in self.arrow_categories]
        for coco_arw in coco_arrows:
            a = id_to_ann[coco_arw["id"]]
            for rel in ARROW_RELATIONS:
                if rel in coco_arw.keys():
                    rel_shape = id_to_ann[coco_arw[rel]]
                    a.set(rel, rel_shape)

        return anns



import numpy as np


def line_segment_lengths(points: np.ndarray) -> np.ndarray:
    """
    Computes the segment lenghts of a polyline represented by a sequence of points
    Args:
        points: point sequence of shape (N,2)

    Returns: segment lengths of shape (N-1,)
    """
    assert points.ndim == 2 and points.shape[1] == 2, f"{points.shape}"
    # the trick here is np.diff, which computes difference between points i and i-1
    return np.sqrt(np.sum(np.diff(points, axis=0) ** 2, axis=1))


def sample_equidistant_points(points: np.ndarray, k: int) -> np.ndarray:
    """
    Resamples a line defined through a sequence of points by generating k equidistant points.
    Args:
        points: polyline of shape (N,2)
        k: the number of equidistant points to sample

    Returns: the sequence of resampled points with shape (k,2)

    """
    segment_lengths = line_segment_lengths(points)
    assert any([l > 0 for l in segment_lengths]), f"Polyline with zero-length: {points}"

    sampled_line_distances = np.linspace(0, np.sum(segment_lengths), k)

    segment_idx = 0
    resampled_pts = [points[0]]

    prev_pt_line_dist = 0
    next_pt_line_dist = segment_lengths[segment_idx]

    # TODO this would be more readable with bisect + list comprehension
    for i in range(1, k - 1):
        line_dist = sampled_line_distances[i]
        # print(f"i={i}, line_dist={line_dist}, segment_idx={segment_idx}, next_pt_line_dist={next_pt_line_dist}")

        # find segment that contains pt specified through line distance
        while line_dist > next_pt_line_dist:
            # print("segment_idx++")
            segment_idx += 1
            prev_pt_line_dist = next_pt_line_dist
            next_pt_line_dist += segment_lengths[segment_idx]

        # calculate factor for segment vector, i.e. where in segment is point located
        ratio = (line_dist - prev_pt_line_dist) / (next_pt_line_dist - prev_pt_line_dist)
        assert 0 < ratio <= 1, f"{ratio}"

        p_i = points[segment_idx]
        p_j = points[segment_idx + 1]
        segment_vect = p_j - p_i

        pt = p_i + ratio * segment_vect
        resampled_pts.append(pt)
        # print(f"p_i={p_i} + ratio={ratio} * (p_j={p_j} - p_i={p_i}) => {pt}")

    resampled_pts.append(points[-1])

    resampled_pts = np.array(resampled_pts)
    return resampled_pts



import numpy as np
import torch

from yamlu import np_utils


def test_bin_stats_np():
    arr = np.array([True, False, True, False, False])
    stat = np_utils.bin_stats(arr)
    assert stat == "2/5 (40.00%)"


def test_bin_stats_np_int():
    arr = np.array([0, 0, 1])
    stat = np_utils.bin_stats(arr)
    assert stat == "1/3 (33.33%)"


def test_bin_stats_torch():
    arr = torch.tensor([False, False, True])
    stat = np_utils.bin_stats(arr)
    assert stat == "1/3 (33.33%)"



import itertools
import json
import logging
import random
from abc import ABC, abstractmethod
from functools import partial
from pathlib import Path
from typing import List, Tuple, Optional, Dict, Union

import numpy as np
from joblib import delayed, Parallel
from joblib.externals import loky
from tqdm import tqdm

from yamlu.img import AnnotatedImage, Annotation
from yamlu.np_utils import to_python_type

_logger = logging.getLogger(__name__)

COCO_FIELD_KEYS = {"id", "category_id", "bbox", "image_id", "iscrowd", "keypoints"}

COCO_ANN_ID_FIELD = "coco_ann_id"


class Dataset(ABC):
    def __init__(
            self,
            dataset_path: Union[Path, str],
            split_n_imgs: Dict[str, int],
            coco_categories: List[Dict],
            keypoint_fields: List[str],
            relation_fields: List[str],
    ):
        """
        :param dataset_path: root of the dataset where images and coco json annotations are saved to
        :param split_n_imgs: number of imgs per split (e.g. {"train": 100, "valid": 5, "test": 5})
        :param coco_categories: coco categories with name, supercategory and id fields
        :param keypoint_fields: name of keypoint fields
        :param relation_fields: relation fields, i.e. name of Annotation fields that point to other Annotation objects
        """
        self.dataset_path = Path(dataset_path) if isinstance(dataset_path, str) else dataset_path
        self.name = self.dataset_path.name

        self.split_n_imgs = split_n_imgs
        self.splits = list(self.split_n_imgs.keys())

        self.coco_categories = coco_categories
        self.cat_name_to_id = {c['name']: c['id'] for c in self.coco_categories}
        self.cat_id_to_name = {c['id']: c['name'] for c in self.coco_categories}

        self.keypoint_fields = keypoint_fields
        self.relation_fields = relation_fields

    def has_keypoints(self):
        return len(self.keypoint_fields) > 0

    @abstractmethod
    def get_split_ann_img(self, split: str, idx: int) -> AnnotatedImage:
        pass


class CocoDatasetExport:
    def __init__(
            self,
            ds: Dataset,
            write_img: bool = True,
            write_ann_img: bool = False,
            sample: Optional[int] = None,
            random_sample: bool = True,
            n_jobs: Optional[int] = None,
            ndigits: int = 3,
    ):
        """
        Args:
            write_ann_img: write image with annotated bounding boxes into an extra directory for debugging purposes
            random_sample: sample images randomly (True) or sample first k (False)
        """
        assert isinstance(write_img, bool)
        assert isinstance(write_ann_img, bool)

        self.write_img = write_img
        self.write_ann_img = write_ann_img

        self.ds = ds
        self.sample = sample
        self.random_sample = random_sample
        self.coco_json_exporter = CocoJsonExporter(self.ds, sample, ndigits)

        self.n_jobs = n_jobs
        if n_jobs is None:
            self.n_jobs = min([loky.backend.context.cpu_count() - 2, 16, max(ds.split_n_imgs.values())])
        _logger.info("Initialized %s with %d jobs", self.__class__.__name__, self.n_jobs)

    def dump_dataset(self) -> Dict[str, List[AnnotatedImage]]:
        return {split: self.dump_split(split) for split in self.ds.splits}

    def dump_split(self, split: str):
        _logger.info("%s: starting split=%s, write_img=%s, write_ann_img=%s, sample=%s, random_sample=%s", self.ds.name,
                     split, self.write_img, self.write_ann_img, self.sample, self.random_sample)
        assert split in self.ds.splits, f"{split} not in {self.ds.splits}"

        split_path = self.create_split_path_dir(split, remove_existing_images=self.write_img)

        ann_imgs_path = split_path.parent / f"{split}_annotated"
        if self.write_ann_img:
            ann_imgs_path.mkdir(exist_ok=True, parents=True)
            # remove existing files in directory
            for p in ann_imgs_path.iterdir():
                p.unlink()

        idxs = list(range(self.ds.split_n_imgs[split]))
        if self.sample is not None and len(idxs) > self.sample:
            if self.random_sample:
                random.seed(0)
                idxs = random.sample(idxs, self.sample)
            else:
                idxs = idxs[:self.sample]

        parallel = Parallel(n_jobs=self.n_jobs)
        ann_imgs = parallel(delayed(self.dump_image)(idx, split, split_path, ann_imgs_path) for idx in tqdm(idxs))

        self.coco_json_exporter.dump_split_coco_json(ann_imgs, split)

        return ann_imgs

    def dump_image(self, idx, split, split_path, ann_imgs_path):
        ann_img = self.ds.get_split_ann_img(split, idx)

        if self.write_img:
            img_path = split_path / ann_img.filename
            ann_img.img.save(img_path)

        if self.write_ann_img:
            ann_img.save_with_anns(ann_imgs_path)

        del ann_img.img

        return ann_img

    def create_split_path_dir(self, split, remove_existing_images: bool):
        ds_path = self.ds.dataset_path
        split_path = ds_path / split if self.sample is None else ds_path / f"sample_{self.sample}" / split

        split_path.mkdir(exist_ok=True, parents=True)

        if remove_existing_images:
            # remove existing files in directory
            for p in split_path.iterdir():
                p.unlink()

        return split_path


class CocoJsonExporter:
    def __init__(self, ds: Dataset, sample: Optional[int], ndigits: int):
        self.ds = ds
        self.sample = check_sample_param(sample)

        self._to_python_type = partial(to_python_type, ndigits=ndigits)

        self.excluded_fields = {COCO_ANN_ID_FIELD, *self.ds.keypoint_fields, *COCO_FIELD_KEYS, *self.ds.relation_fields}

    def dump_split_coco_json(self, ann_imgs: List[AnnotatedImage], split: str):
        indent = None if self.sample is None else 2

        coco_json_path = self._split_coco_json_path(split)
        _logger.info("Start dumping coco %d annotations to %s", len(ann_imgs), coco_json_path)
        coco = self.create_coco_dict(ann_imgs)
        _logger.info("Finished creating coco annotations")

        coco_json_path.parent.mkdir(exist_ok=True, parents=True)
        try:
            with coco_json_path.open("w") as f:
                json.dump(coco, f, indent=indent)
        except TypeError as e:
            _logger.warning("Coco object not json serializable:\n%s", coco)
            raise e

    def create_coco_dict(self, ann_imgs: List[AnnotatedImage]) -> Dict:
        coco = {
            "images": _create_images_dict(ann_imgs),
            "annotations": self._create_annotations_dict(ann_imgs),
            "categories": self.ds.coco_categories
        }
        return coco

    def _split_coco_json_path(self, split: str) -> Path:
        if self.sample is None:
            return self.ds.dataset_path / f"{split}.json"
        else:
            return self.ds.dataset_path / f"sample_{self.sample}" / f"{split}.json"

    def _create_img_anns(self, t: Tuple[int, AnnotatedImage]) -> List[Dict]:
        img_id, ann_img = t

        # create artifical unique annotation id
        # assuming that an image has less than 1000 annotations
        assert len(ann_img.annotations) < 1000

        # set id field first so that they can be used for relation mappings
        for i, ann in enumerate(ann_img.annotations):
            ann.set(COCO_ANN_ID_FIELD, int(img_id * 1000 + i))

        coco_anns = [self._create_coco_ann(ann, img_id) for ann in ann_img.annotations]
        return coco_anns

    def _create_coco_ann(self, ann: Annotation, img_id: int) -> Dict:

        coco_ann = {
            "id": ann.get(COCO_ANN_ID_FIELD),
            "image_id": int(img_id),
            "category": ann.category,  # this isn't strictly required but makes debugging easier
            "category_id": self.ds.cat_name_to_id[ann.category],
            "area": self._to_python_type(ann.bb.area),  # needed by coco eval
            "bbox": list(map(self._to_python_type, ann.bb.bb_coco)),
            "iscrowd": 0,
        }

        if self.ds.has_keypoints():
            # in coco format convention keypoints have an additional visibility flag
            # - v=0: not labeled (in which case x=y=0)
            # - v=1: labeled but not visible
            # - v=2: labeled and visible
            # see: http://cocodataset.org/#format-data
            # this means each keypoint is represented as [x,y,v]

            # frameworks like detectron2 assume that if the dataset has keypoints, this applies to all categories
            # as a workaround we create v=0 keypoints for categories without keypoints
            kps = np.zeros(3 * len(self.ds.keypoint_fields))

            for i, k in enumerate(self.ds.keypoint_fields):
                if k in ann:
                    x, y = getattr(ann, k)
                    kps[i * 3: i * 3 + 3] = [x, y, 2]

            coco_ann["keypoints"] = self._to_python_type(kps)

        for rel_name in self.ds.relation_fields:
            if rel_name in ann:
                # relationship fields point to other Annotation objects
                rel_ann = ann.get(rel_name)
                coco_ann[rel_name] = rel_ann.get(COCO_ANN_ID_FIELD)

        # make sure we don't overwrite reserved fields
        coco_ann.update({
            k: self._to_python_type(v) for k, v in ann.extra_fields.items() if
            k not in self.excluded_fields and not isinstance(v, Annotation)
        })
        return coco_ann

    def _create_annotations_dict(self, ann_imgs: List[AnnotatedImage]):
        imgs_anns = [self._create_img_anns((i, ai)) for i, ai in enumerate(tqdm(ann_imgs))]
        flat_anns = list(itertools.chain(*imgs_anns))
        return flat_anns


def check_sample_param(sample):
    assert sample is None or not isinstance(sample, bool), f"sample is not None or an int: {sample} ({type(sample)})"
    return sample


def _create_images_dict(ann_imgs: List[AnnotatedImage]):
    return [
        {
            "file_name": ann_img.filename,
            "height": int(ann_img.height),
            "width": int(ann_img.width),
            "id": idx
        } for idx, ann_img in enumerate(ann_imgs)
    ]



from typing import List, Union

import numpy as np
from scipy.sparse.csgraph import connected_components

from yamlu.img import Annotation

# reading directions for how to combine words into a textblock:
# left-to-right (lr), right-to-left (rl), top-to-bottom (tb), bottom-to-top (bt)
DIRECTIONS = ["lr", "rl", "tb", "bt"]


def transcribe_words(word_anns: List[Annotation], word_idxs: Union[np.array, List], line_thresh: float, line_delim="\n",
                     direction: str = "lr") -> str:
    """
    Transcribe a text block given words and their bounding boxes

    Args:
        word_anns: word annotations with the text in the category attribute
        word_idxs: the annotation indices to transcribe
        line_thresh: max y_mid dist in px such that two word bounding boxes are still considered to be on the same line
        line_delim: delimiter character for multiple lines
        direction: reading direction, one of lr, rl, tb, bt

    Returns:
        the transcribed text block as string

    """
    assert direction in DIRECTIONS, f"Invalid direction: {direction}"
    xc, yc = np.array([word_anns[i].bb.center for i in word_idxs]).T
    word_idxs = np.asanyarray(word_idxs)

    line_split_axis = yc if direction in ["lr", "rl"] else xc
    word_split_axis = xc if direction in ["lr", "rl"] else yc

    adj_matrix = np.abs(line_split_axis[np.newaxis, :] - line_split_axis[:, np.newaxis]) <= line_thresh
    n_components, labels = connected_components(adj_matrix, directed=False)

    line_means = np.array([line_split_axis[labels == i].mean() for i in range(n_components)])
    labels_ordered = line_means.argsort() if direction in ["lr", "bt"] else line_means.argsort()[::-1]

    lines = []
    for i in labels_ordered:
        word_mask_line = labels == i
        word_idxs_line = word_idxs[word_mask_line]
        word_order = word_split_axis[word_mask_line].argsort()
        word_order = word_order if direction in ["lr", "tb"] else np.flip(word_order, [0])
        sorted_word_idxs = word_idxs_line[word_order]
        if not len(sorted_word_idxs.shape):
            sorted_word_idxs = sorted_word_idxs[..., np.newaxis]
        line = " ".join([word_anns[wi].category for wi in sorted_word_idxs])
        lines.append(line)

    return line_delim.join(lines)
